/*
 * Crypt.java
 *
 * Crypt or proliferative unit.
 */

package sim;
import java.util.ArrayList;
import java.awt.Color;
import lib.KnuthRandom;
import lib.Probability;

public class Crypt {
    // Data Members
    private eCryptState _state = eCryptState.AliveWT; // current state of crypt
    protected PGrid _grid;      // grid that manages the geometry of crypts
    protected int _nVertexIndex;// position in Grid class
    protected ArrayList<Crypt> _neighbors; // list of neighboring crypts
    private Crypt _parent=null;      // reference to the crypt's parent
    private Crypt _cloneParent=null; // reference to the mutant crypt that started the clone to which this crypt belongs (only if it's mutant)

    private double _mutationRate=0.0;          // mutation rate
    private double _meanCryptDivisionTime=2.0; // 
    private double _wtBackgroundDeathRate=0.0; // background death rate
    private double _mutantSurvivalAdvantageFactor=1.0;
    private double _mutantWoundResistanceFactor=1.0;
    private double _mutantReproductiveAdvantageFactor=1.0;

    private Color  _color;                 // color of crypt for display

    // For simulation event queue management
    SimulationEvent _eventDeath = null;
    SimulationEvent _eventMutate = null;
    SimulationEvent _eventDivide = null;

    // Class constants
    private static final Color COLOR_DEAD = Color.black;     // color for dead crypts
    private static final Color COLOR_ALIVE_WT = Color.white; // color for wt crypts

    // Enum
    public enum eCryptState { Dead, AliveWT, AliveMutant };
    
    // Constructor
    public Crypt() { 
	_neighbors = new ArrayList<Crypt>();
	_color = COLOR_ALIVE_WT;
    }

    // Copy Constructor
    public Crypt( Crypt rhs ) {
	_nVertexIndex = rhs._nVertexIndex;
	_grid = rhs._grid;
    }

    public void setParams(TissueEnvironmentSettings params) {
	setMeanCryptDivisionTime(params.get_meanCryptDivisionTime());
	setMutantSurvivalAdvantageFactor(params.get_mutantSurvivalAdvantageFactor());
	setMutantWoundResistanceFactor(params.get_mutantWoundResistanceFactor());
	setMutantReproductiveAdvantageFactor(params.get_mutantReproductiveAdvantageFactor());
	setMutationRate(params.get_mutationRate());
	setWtBackgroundDeathRate(params.get_wtBackgroundDeathRate());
    }

    public void setMeanCryptDivisionTime(double x) { _meanCryptDivisionTime=x; }
    public void setMutantReproductiveAdvantageFactor(double x) { _mutantReproductiveAdvantageFactor=x; }
    public void setMutantSurvivalAdvantageFactor(double x) { _mutantSurvivalAdvantageFactor=x; }
    public void setMutantWoundResistanceFactor(double x) { _mutantWoundResistanceFactor=x; }
    public void setMutationRate(double x) { _mutationRate=x; }
    public void setWtBackgroundDeathRate(double x) { _wtBackgroundDeathRate=x; }
	
    public void addNeighbor(Crypt c) { _neighbors.add(c); }
    public ArrayList<Crypt> getNeighbors() { return _neighbors; }
    public void setGrid(PGrid g) { _grid = g; }
    public void setIndex(int n) { _nVertexIndex = n; }
    public int getIndex() { return _nVertexIndex; }
    public double getX() { return _grid.getX(_nVertexIndex); }
    public double getY() { return _grid.getY(_nVertexIndex); }
    public Color getColor() { return _color; }
    public void setColor(Color color) { _color = color; }
	
    public boolean isAlive() { return (_state != eCryptState.Dead); }
    public eCryptState GetCryptState() {
	return _state;
    }
    public void SetCryptState(eCryptState state) {
	_state = state;
    }
    public Crypt getCloneParent() { return _cloneParent; }
    public void setCloneParent(Crypt parent) { _cloneParent = parent; }
    public Crypt getParent() { return _parent; }
    public void setParent(Crypt parent) { _parent = parent; }

    // Computes the distance from this crypt to the specified location on the tissue
    public double distanceTo( double x, double y )
    {
	return Math.hypot( (getX()-x), (getY()-y) );
    }

    // Returns a list of the ordinal index of all neighbors with the specified state of being alive (true/false)
    protected ArrayList<Integer> GetLivingNeighborIndices()
    {
	ArrayList<Integer> neighborIndices = new ArrayList<Integer>();
	for( int i = 0; i < _neighbors.size(); i++ ) {
	    Crypt neighbor = _neighbors.get(i);
	    if (neighbor.isAlive())
		neighborIndices.add( new Integer(i) );
	}
		
	return neighborIndices;
    }

    protected ArrayList<Integer> GetDeadNeighborIndices()
    {
	ArrayList<Integer> neighborIndices = new ArrayList<Integer>();
	for( int i = 0; i < _neighbors.size(); i++ ) {
	    Crypt neighbor = _neighbors.get(i);
	    if (!neighbor.isAlive())
		neighborIndices.add( new Integer(i) );
	}
		
	return neighborIndices;
    }

    public void die(Scheduler s)
	throws Exception
    {
	// Change state/color to reflect death
	_state = eCryptState.Dead;
	_color = Crypt.COLOR_DEAD;

	// Invalidate remaining events
	if( _eventDeath != null ) _eventDeath.Invalidate();
	if( _eventMutate != null ) _eventMutate.Invalidate();
	if( _eventDivide != null ) _eventDivide.Invalidate();
	_eventDeath=null;
	_eventMutate=null;
	_eventDivide=null;
	
	// Send the death announcement to all my neighbors
	ArrayList<Integer> livingNeighborIndices = GetLivingNeighborIndices();
	for(int i=0; i<livingNeighborIndices.size(); i++ ) {
	    Crypt targetCrypt = _neighbors.get( livingNeighborIndices.get(i) );
	    targetCrypt.NeighborDeathSignal(s);
	}
    }

    private int GetRandomNeighborIndex(Scheduler s, ArrayList<Integer> neighborIndices )
    {
	int iIndex = -1;
	if( neighborIndices.size() > 0 ) {
	    int iRandNeighbor = s.GetRNG().randomInt(neighborIndices.size());	
	    iIndex = neighborIndices.get( iRandNeighbor ).intValue();
	}		
	return iIndex;
    }

    public void divide(Scheduler s)
	throws Exception
    {
	// Get indexes of all dead neighbors
	ArrayList<Integer> alDeadNeighborIndices = GetDeadNeighborIndices();
		
	// Randomly select one
	if( alDeadNeighborIndices.size() > 0 ) {
			
	    int iRandNeighborIndex = GetRandomNeighborIndex(s, alDeadNeighborIndices );
			
	    // Take over that location in the tissue with a daughter crypt
	    // TODO this is reusing the Crypt object - do I need to perform a reset?
	    Crypt daughter = _neighbors.get( iRandNeighborIndex );

	    daughter.SetCryptState( this._state );
	    daughter.setParent(this);
	    daughter.setCloneParent(getCloneParent());
	    daughter.setColor( this.getColor());

	    // Generate event for new daughter crypt's death
	    SimulationEvent eventDaughterDeath = daughter.GenerateDeathEvent(s);
	    s.PostEvent(eventDaughterDeath);

	    // If daughter is wild-type, generate event for daughter crypt's mutation
	    if( daughter.GetCryptState() == eCryptState.AliveWT ) {
		SimulationEvent eventDaughterMutation = daughter.GenerateMutationEvent(s);
		s.PostEvent(eventDaughterMutation);
	    }
			
	    // If daughter has any dead neighbors, generate event for daughter crypt's division
	    if( daughter.GetDeadNeighborIndices().size() > 0 ) {
		SimulationEvent eventDaughterDivide = daughter.GenerateNextDivisionEvent(s);
		s.PostEvent(eventDaughterDivide);
	    }

	    // Check if the parent crypt has any remaining dead neighbors; if so, generate next division event
	    // We can reuse the array we already computed;  we already occupied one empty neighbor position, are there more?
	    if( alDeadNeighborIndices.size() > 1 ) {
		SimulationEvent eventParentDivide = GenerateNextDivisionEvent(s);
		s.PostEvent(eventParentDivide);
	    }
	}
    }
    
    public void generateCloneColor(KnuthRandom r) {
	// Note the restriction of the color components to [2..253] to avoid black and white
	_color = new Color( r.randomInt(251)+2, r.randomInt(251)+2, r.randomInt(251)+2);
    }

    // mutate - turns this crypt into a living mutant, even if the crypt was dead
    public void mutate(Scheduler s)
	throws Exception
    {
	_state = eCryptState.AliveMutant; // note that a dead crypt will come to life
		
	// We are the parent of this brand new clone
	setCloneParent(this);

	// Generate a random color for this clone
	generateCloneColor(s.GetRNG());

	// do mutants have different death and division rates?
	s.PostEvent(GenerateDeathEvent(s));
	if( GetDeadNeighborIndices().size() > 0 )
	    s.PostEvent(GenerateNextDivisionEvent(s));
    }
	
    // This is invoked when one of 'this's neighbors died; we need to 
    //  queue up a division event to attempt to occupy the vacated area
    public void NeighborDeathSignal(Scheduler s)
	throws Exception
    {
	SimulationEvent event = this.GenerateNextDivisionEvent(s);
	s.PostEvent(event);	// Queue the division event
    }
	
    /*
     * EVENT GENERATION
     */
    
    public SimulationEvent GenerateDeathEvent(Scheduler s) {
	double effectiveAdvantageFactor = (this.GetCryptState() == eCryptState.AliveWT ? 1.0 :_mutantSurvivalAdvantageFactor );
	if( _eventDeath != null )
	    _eventDeath.Invalidate();
	if (_wtBackgroundDeathRate * effectiveAdvantageFactor>0.0) {
	    double ttDeath = s.GetCurrentTime() + Probability.RandomFromExponential(_wtBackgroundDeathRate * effectiveAdvantageFactor, s.GetRNG());
	    _eventDeath = new SimulationEvent( this, ttDeath, SimulationEvent.EventType.death );
	} else {
	    _eventDeath = null;
	}
	return _eventDeath;
    }
	
    public SimulationEvent GenerateMutationEvent(Scheduler s) {
	if( _eventMutate != null )
	    _eventMutate.Invalidate();
	if (_mutationRate>0.0) {
	    double ttMutation = s.GetCurrentTime() + Probability.RandomFromExponential(_mutationRate, s.GetRNG());
	    _eventMutate = new SimulationEvent( this, ttMutation, SimulationEvent.EventType.mutation );
	} else {
	    _eventMutate = null;
	}
	return _eventMutate;
    }
    
    public SimulationEvent GenerateNextDivisionEvent(Scheduler s) {
	double meantime = _meanCryptDivisionTime * (_state == eCryptState.AliveMutant?_mutantReproductiveAdvantageFactor:1.0);
	double ttDivision = s.GetCurrentTime() + (meantime>0.0?(Probability.RandomFromExponential(1.0/meantime, s.GetRNG())):0.0);
	if( _eventDivide != null)
	    // if the crypt was about to divide anyways, don't create a new division event
	    if (_eventDivide.isValid() &&_eventDivide.get_time()<=ttDivision && _eventDivide.get_time()>s.GetCurrentTime())
		return _eventDivide;
	    else
		_eventDivide.Invalidate();
	_eventDivide = new SimulationEvent( this, ttDivision, SimulationEvent.EventType.division );		
	return _eventDivide;
    }
}
