/*
 * Scheduler.java
 * event queue management for Crypts on a PGrid
 */

package sim;

import java.awt.*;
import java.io.*;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.*;

import sim.Crypt.eCryptState;
import lib.KnuthRandom;
import sim.SimulationEvent.EventType;

public class Scheduler {
    // Instance members
    protected KnuthRandom _r = null;
    protected double _fCurrentDay = 0.0;  // current time in days
    protected PriorityQueue<SimulationEvent> _priorityQueue;
    protected TissueEnvironmentSettings _envSettings;
    protected boolean _bCheckTissueDominance;
    protected PGrid _grid;

    // Enums
    public enum SimulationTerminationReason { Unknown, MaxTimeReached, MaxEventsProcessed, DominatedByMutants, DominatedByDeath, NoMutants, EmptyQueue, Exception };
	
    // Constructor
    public Scheduler( PGrid grid, long iRandomSeed, TissueEnvironmentSettings envSettings )
	throws Exception
    {
	_r = new KnuthRandom();
	_r.seedRandom(iRandomSeed);
	_grid = grid;
	_bCheckTissueDominance = false;
		
	// Create the empty priority queue
	_priorityQueue = new PriorityQueue<SimulationEvent>( 100000, new SimulationEvent.SimulationEventComparator() );	
	_envSettings = envSettings;

	// Setup the priority queue and get initial events from the cells
	InitializePQueue();
    }
	
    /*
     * INITIALIZATION
     */
    protected void InitializePQueue()
	throws Exception
    {
	int numUnits = _grid.getSize();
		
	// Populate the queue with the initial events for each unit
	for( int i=0; i<numUnits; i++ ) {
	    Crypt crypt = _grid.getCrypt(i);
			
	    // Create and post this unit's death event
	    SimulationEvent simEventDeath = crypt.GenerateDeathEvent(this);
	    if (simEventDeath!=null)
		PostEvent( simEventDeath );
	    
	    // Create and post this unit's mutation event
	    SimulationEvent simEventMutation = crypt.GenerateMutationEvent(this);
	    if (simEventMutation!=null)
		PostEvent( simEventMutation );
	}
		
	// The maximum number of days to run for
	final long maximumDays = _envSettings.get_maxSimulationYears() * 365;
	PostEvent(new SimulationEvent(null, maximumDays, SimulationEvent.EventType.endSimulation));

	// Post wounding events
	if (_envSettings.get_woundingMode()!=TissueEnvironmentSettings.WoundingMode.None && _envSettings.get_woundIntervalTime()>0.0) {
	    for (double i=_envSettings.get_woundIntervalTime(); i<maximumDays; i+=_envSettings.get_woundIntervalTime()) {
		if(_envSettings.get_woundingMode() == TissueEnvironmentSettings.WoundingMode.Localized)
		    PostEvent(new SimulationEvent(_grid.getCryptAtCenter(), i, _envSettings.get_woundRadius(), SimulationEvent.EventType.wound ));
		else
		    PostEvent(new SimulationEvent(null, i, 0, SimulationEvent.EventType.wound ));
	    }
	}
    }

    // Should the event loop check for tissue dominance?
    public void SetTissueDominanceCheck(boolean b) {
	_bCheckTissueDominance = b;
    }

    /*
     * Accessors
     */
    public PGrid GetGrid() { return _grid; }
    public double GetCurrentTime() { return _fCurrentDay; }
    public KnuthRandom GetRNG() { return _r; }
    
    /*    public TissueEnvironmentSettings get_envSettings() {
	  return _envSettings;
	  }	*/

    protected SimulationEvent GetNextEvent()
	throws Exception
    {	
	// Pluck the next event from the queue; the poll() call is synchronous
	//  and returns null if there is not a pending event
	SimulationEvent curEvent = _priorityQueue.poll();
	if( curEvent.get_time() < _fCurrentDay )
	    throw new Exception( "TissueBase:GetNextEvent - event is earlier than current time");
	return curEvent;
    }
	
    // Appends event to end of the queue 
    public boolean PostEvent( SimulationEvent event ) 
	throws Exception
    {	
	if (event==null)
	    return false;

	// Prevent adding an event that happens in the past
	//	if( event.get_time() <= this._fCurrentDay )
	if( event.get_time() < _fCurrentDay )
	    throw new Exception( "TissueBase:PostEvent - cannot schedule event in the past");
		
	// Prevent re-posting the same event
	//	if( event.isPostedToQueue() ) {
	//	    _priorityQueue.remove(event);
	//	    assert false;
	    // Note - a bug in the PriorityQueue implementation before jse1.6 may cause problems
	//	}
		
	// Add the event to the queue; if successful, mark it as having been added
	if( !_priorityQueue.add( event ) )
	    throw new Exception( "Error posting event to queue " + event);
	
	return true;
    }

    /*
     * MAIN SIMULATION LOOP
     */
	
    // The optional 'progressUpdater' parameter allows the caller to receive
    // periodic notifications of progress.
    public SimulationTerminationReason RunSimulation(ISimulationProgressCallback progressUpdater) 
    {
	// How many events were processed since the last progress callback
	long numEventsProcessedSinceLastProgressUpdate = 0;
	
	// How many events have we processed so far
	long numTotalEventsProcessed = 0;

	// For tissue dominance checks
	//	final double totalCrypts = (double)_grid.getSize();
	boolean bTissueDominated = false;
	//	long nextTissueDominanceCheck = _kTissueDominanceCheckInterval;
		
	// Why did this simulation terminate?
	SimulationTerminationReason eTermReason = SimulationTerminationReason.Unknown;
		
	// Setup progress callbacks
	int progressUpdateInterval = (progressUpdater != null ? progressUpdater.get_progressInterval() : Integer.MAX_VALUE);
		
	try {
	    SimulationEvent curEvent;
	    do {
		curEvent = GetNextEvent();
		if (_fCurrentDay > curEvent.get_time()) {
		    System.err.println("Simulation:RunSimulation: Possible PriorityQueue bug! Events out of order! "+curEvent.get_time()+curEvent.isValid());
		    System.out.println("Simulation:RunSimulation: Fatal error at time "+_fCurrentDay);
		    System.exit(-1);
		}

		_fCurrentDay = curEvent.get_time();
		
		// Continue until event queue is empty
		// Only process this event if it is valid (i.e. hasn't been
		// marked 'deleted')
		if ( curEvent != null && curEvent.isValid()) {
		    ++numEventsProcessedSinceLastProgressUpdate;
		    ++numTotalEventsProcessed;
					
		    // Is it time to make a progress update callback?
		    if( numEventsProcessedSinceLastProgressUpdate >=
			progressUpdateInterval ){
			progressUpdater.Execute(new Long(numTotalEventsProcessed));
			numEventsProcessedSinceLastProgressUpdate = 0;
			progressUpdateInterval = (progressUpdater != null ?
						  progressUpdater.get_progressInterval() : Integer.MAX_VALUE);
		    }
					
		    switch (curEvent.get_eventType()) {
		    case callback:
			if (progressUpdater != null)
			    progressUpdater.Execute(new Long(numTotalEventsProcessed));
			numEventsProcessedSinceLastProgressUpdate = 0;
			progressUpdateInterval = (progressUpdater != null ? progressUpdater.get_progressInterval() : Integer.MAX_VALUE);
			break;
		    case wound:
			InflictAcuteWound(curEvent);
			break;
		    case death:
			curEvent.get_cryptRef().die(this);
			break;
		    case division:
			curEvent.get_cryptRef().divide(this);
			break;
		    case mutation:
			curEvent.get_cryptRef().mutate(this);
			break;
		    case seedMutant:
			PlantMutantSeed();
			break;
		    case endSimulation:
			break;
		    default:
			System.err.println("ERROR: Scheduler: unknown event type discarded");
		    }
		    // Invalidate this event
		    curEvent.Invalidate();
		}
		if (_bCheckTissueDominance) {
		    // Get counts of all cell states at once
		    int[] stateCounts = _grid.CountCryptStates();
		}
		/*				
		// Check whether we need to check whether tissue domination has
		// occurred
		if( numTotalEventsProcessed >= nextTissueDominanceCheck ) {
					
		double pctDead = 0.0;
					
		// Get counts of all cell states at once
		int[] stateCounts = _grid.CountCryptStates();
					
		// Check whether mutants or dead cells dominate the tissue
		if( (pctDead = ((double)stateCounts[Crypt.eCryptState.Dead.ordinal()]/totalCrypts)) > Constants.TISSUEDEATHDOMINANCETHRESHOLD ) {
		bTissueDominated = true;
		eTermReason = SimulationTerminationReason.DominatedByDeath;
		}
		else if( (double)stateCounts[Crypt.eCryptState.AliveMutant.ordinal()]/totalCrypts > Constants.TISSUEMUTANTDOMINANCETHRESHOLD ) {
		bTissueDominated = true;
		eTermReason = SimulationTerminationReason.DominatedByMutants;
		}
		// 02/17/2006 ABORT IF NO MUTANTS
		// JTE TODO - this type of aborting should only happen when
		// we are performing clonal tests
		// in which we are investigating null mutation rate
		else if( _bPlantedMutantSeed && stateCounts[Crypt.eCryptState.AliveMutant.ordinal()] == 0 ) {
		bTissueDominated = true;
		eTermReason = SimulationTerminationReason.NoMutants;	
		}
		// 02/26/2006 PLANT MUTANT SEED
		else if ( !_bPlantedMutantSeed && pctDead>0.001) {
		PlantMutantSeed(1);
		// Schedule next dominance check
		nextTissueDominanceCheck += _kTissueDominanceCheckInterval;
						
		// JTE TODO - shouldn't be duplicating this code;
		// cleanup this whole section
		}
					
		else {
		// Schedule next dominance check
		nextTissueDominanceCheck += _kTissueDominanceCheckInterval;
		}
		}
		*/
	    } while (curEvent.get_eventType()!=SimulationEvent.EventType.endSimulation && !bTissueDominated);		
	    // Determine cause of loop termination (other than tissue dominance
	    // which was already set)
	    if( curEvent == null )
		eTermReason = SimulationTerminationReason.EmptyQueue;
	    else if (curEvent.get_eventType()==SimulationEvent.EventType.endSimulation)
		eTermReason = SimulationTerminationReason.MaxTimeReached;
	    // DumpCellStates();
	    // DumpPQueue();
	}
	catch( Exception ex ) {
	    System.err.println("RunSimulation caught exception " + ex.getMessage() );
	    ex.printStackTrace(System.err);
	    eTermReason = SimulationTerminationReason.Exception;
	}
	//	System.err.println("Done due to " + eTermReason);
	return eTermReason;
    }

    protected void InflictAcuteWound(SimulationEvent event) throws Exception
    {
	// wounds events with a crypt are local and centered on the crypt
	if (event.get_cryptRef()!=null) {
	    ArrayList<Crypt> cryptsWithinWound = _grid.getCryptsInCircle(event.get_cryptRef().getX(), event.get_cryptRef().getY(), event.get_param());
	    for( Crypt curCrypt : cryptsWithinWound ) {
		if( curCrypt.isAlive() ) {
		    // Does this crypt have a survival advantage?
		    double effectiveAdvantageFactor = (curCrypt.GetCryptState() == eCryptState.AliveWT ? 1.0 : _envSettings.get_mutantWoundResistanceFactor() );
		    
		    // Check whether this crypt is killed...
		    if( _r.randomDouble() < effectiveAdvantageFactor * _envSettings.get_wtWoundImpactProbability() ) {
			curCrypt.die(this);
		    }
		}
	    }
	} else {	// wounds events with no crypt are distributed

	    for (int i=0; i<_grid.getSize(); i++) {
		Crypt curCrypt = _grid.getCrypt(i);
		if( curCrypt.isAlive() ) {
		    // Does this crypt have a survival advantage?
		    double effectiveAdvantageFactor = (curCrypt.GetCryptState() == eCryptState.AliveWT ? 1.0 : _envSettings.get_mutantWoundResistanceFactor() );
		    
		    // Check whether this crypt is killed...
		    if( _r.randomDouble() < effectiveAdvantageFactor * _envSettings.get_wtWoundImpactProbability() ) {
			curCrypt.die(this);
		    }
		}
	    }

	}
    }

    // Plants a seed of mutant(s) at the center of the tissue
    // For use in clonal testing in which mutations are disabled
    protected void PlantMutantSeed() throws Exception
    {
	Crypt crypt = _grid.getCryptAtCenter();
	crypt.mutate(this);  // what if it is dead???????????????
    }

    /*
     * INPUT/OUTPUT
     */
    /*
      public void DumpCellStates() {
	
      int numCrypts = _grid.getSize();
      for (int i=0; i<numCrypts; i++) {
		
      // Get the current crypt
      // jte -- note downcast
      Crypt crypt = (Crypt)_grid.getCrypt(i);
      NullCryptCells cells = (NullCryptCells )crypt._cells;
      System.out.println(cells);
      }
      }
    */


    public void SaveState(File destination)
	throws IOException
    {		
	PrintWriter output = null;
	try {
	    output = new PrintWriter( new BufferedWriter( new FileWriter(destination) ) );
		
	    StringBuffer sb = new StringBuffer();
	    double currentRow = 0.0;
		
	    // foreach crypt...
	    int numCrypts = _grid.getSize();
	    for (int i=0; i<numCrypts; i++) {
		// Get the current crypt
		Crypt crypt = _grid.getCrypt(i);
		if (crypt!=null) {
		    // Are we starting a new row?
		    if( crypt.getY() != currentRow )
			{
			    output.println( sb.toString() );
					
			    currentRow = crypt.getY();
			    sb = new StringBuffer();
			}
				
		    // Map the crypt state to an int
		    int intState = 0;
		    Crypt.eCryptState state = crypt.GetCryptState();
		    if( state == Crypt.eCryptState.AliveWT )
			intState = 1;
		    else if( state == Crypt.eCryptState.AliveMutant )
			intState = 2;
		    // default of 0 is already set for .Dead
				
		    // Append the current crypt's state to the StringBuffer
		    sb.append( intState ).append( " " );
		}
	    }			
		
	}
	finally {
	    if (output != null) output.close();
	}	
    }

    /*  public void UpdateStatistics( TissueEnvironmentSettings params, SimulationTerminationReason eTermReason, double lNumTimestamps, String sStateFilename, String sImageFilename, long elapsedTimeSec, BasicStatsSet stats, FractalGeometry.FractalGeometryStats fractalStats, double dPctDead, double dPctWT, double dPctMutant, FragmentationStatsSet fragmentationStats, File statisticsFile )
	throws IOException
	{
	NumberFormat longDecimalFormatter = new DecimalFormat("#0.0000");
	NumberFormat shortDecimalFormatter = new DecimalFormat("#0");
	
	PrintWriter out =  new PrintWriter(( new BufferedWriter(new FileWriter( statisticsFile, true))));        
	StringBuffer sb = new StringBuffer();		
	
	// Label, runtime, output files
	sb.append( params.get_label() ).append("\t").append(eTermReason).append("\t").append( elapsedTimeSec ).append("\t").append( ((Double)lNumTimestamps).intValue() ).append("\t").append( sStateFilename ).append("\t").append( sImageFilename ).append("\t");
	
	// Parameters of the run
	sb.append( params.get_mutationProbability()).append("\t").append( params.get_wtBackgroundDeathProbability() ).append("\t");
	sb.append( params.get_wtWoundImpactProbability() ).append("\t").append( params.get_woundIntervalTime() ).append("\t").append( params.get_mutantSurvivalAdvantageFactor()).append("\t").append( params.get_mutantReproductiveAdvantageFactor()).append("\t");
	sb.append( params.get_woundingMode()).append("\t").append(params.get_maxSimulationYears()).append("\t");
	
	// Statistics on the run
	sb.append( stats.Count ).append("\t").append( stats.Min ).append("\t").append( stats.Max ).append("\t");
	sb.append( shortDecimalFormatter.format(stats.Mean) ).append("\t").append( stats.Median).append("\t").append( longDecimalFormatter.format(stats.StdDev)).append("\t");
	sb.append( dPctDead).append("\t").append(dPctWT).append("\t").append(dPctMutant).append("\t");
	
	sb.append( longDecimalFormatter.format(fractalStats.DimensionCoarse)).append("\t").append(longDecimalFormatter.format(fractalStats.DimensionFine)).append("\t").append(longDecimalFormatter.format(fractalStats.LacunarityCoarse)).append("\t").append(longDecimalFormatter.format(fractalStats.LacunarityFine)).append("\t");
	
	// Distribution of clone sizes
	sb.append(stats.CloneSizeDistribution).append( "\t" );
	
	// Distribution of significant fragment sizes
	sb.append(fragmentationStats.get_FragmentSizeStats()).append( "\t" );
	
	out.println( sb.toString() );
	out.close();
	}
    */
}
